import numpy as np
import pandas as pd
from pandas import DataFrame
from sklearn.base import BaseEstimator
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.linear_model import LogisticRegression, Lasso, LinearRegression
from sklearn.model_selection import train_test_split
from sklearn.model_selection._search import BaseSearchCV, GridSearchCV, RandomizedSearchCV
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from xgboost import XGBClassifier, XGBRegressor

from adaptee import HyperBRKGASearchCV


def get_cv_time(method: BaseSearchCV):
    mean_fit_time = method.cv_results_['mean_fit_time']
    mean_score_time = method.cv_results_['mean_score_time']
    n_splits = method.n_splits_  # number of splits of training data
    n_iter = pd.DataFrame(method.cv_results_).shape[0]  # Iterations per split
    return np.mean(mean_fit_time + mean_score_time) * n_splits * n_iter


def get_metrics(filename: str, hyperopt_method, is_hyperbrkga: bool, df: DataFrame):
    if is_hyperbrkga:
        score = hyperopt_method.cv_results_['best_param_score']
        time = hyperopt_method.cv_results_['total_time']
        params = str(hyperopt_method.cv_results_['best_param_decoded'])
    else:
        score = hyperopt_method.best_score_
        time = get_cv_time(hyperopt_method)
        params = str(hyperopt_method.best_params_)

    row = {
        'Score': score,
        'Time': time,
        'Param': params,
    }

    df = df.append(row, ignore_index=True)
    return df, row


def compute_mean_and_std(df, filename):
    means = {
        'Score': '',
        'Time': '',
        'Param': '',
    }
    for i in df.columns:
        if 'Param' not in i:
            means[i] = f'Média: {df[i].mean()}'

    std = {'Score': f'Desvio Padrão: {df["Score"].std()}'}

    new_df = DataFrame()
    new_df = new_df.append(means, ignore_index=True)
    new_df = new_df.append(std, ignore_index=True)

    new_df.to_csv(filename, mode='a')


def cred_preprocessing():
    # Definição dos nomes das variáveis (conforme a tabela contida no enunciado)
    colnames = ['ESCT', 'NDEP', 'RENDA', 'TIPOR', 'VBEM', 'NPARC',
                'VPARC', 'TEL', 'IDADE', 'RESMS', 'ENTRADA', 'CLASSE']
    # Leitura dos dados de treino
    arquivo = './experiments/credtrain.txt'
    data_train = pd.read_csv(arquivo, sep='\t', header=None, names=colnames)
    # Leitura dos dados de teste
    arquivo = './experiments/credtest.txt'
    data_test = pd.read_csv(arquivo, sep='\t', header=None, names=colnames)
    # Aplicação no conjunto de treinamento
    data_train_new = pd.get_dummies(data=data_train,
                                    prefix='ESCT',
                                    columns=['ESCT'],
                                    drop_first=True)
    data_test_new = pd.get_dummies(data=data_test, prefix='ESCT', columns=['ESCT'], drop_first=True)
    # Transformação da variável alvo do conjunto de treinamento e teste em vetor
    y_train = np.array(data_train_new['CLASSE'])
    y_test = np.array(data_test_new['CLASSE'])
    features_name_train = list(data_train_new.columns)  # nomes das colunas
    features_name_train.remove('CLASSE')  # remove variável "CLASSE"
    X_train = np.array(data_train_new.loc[:, features_name_train])  # Transformação em matriz de dados
    X_train = X_train.astype(float)
    # Transformação do conjunto de teste remanescente em matriz de dados
    features_name_test = list(data_test_new.columns)  # Nomes das colunas
    features_name_test.remove('CLASSE')  # Remoção da variável "CLASSE"
    X_test = np.array(data_test_new.loc[:, features_name_test])  # Transformação em matriz
    X_test = X_test.astype(float)
    # Criação do objeto para a padronização das features
    scaler = StandardScaler()
    # Ajustamento do StandardScaler ao conjunto de dados de treino e padronização dos dados de treino
    X_train = scaler.fit_transform(X_train)
    # Transformação dos dados de teste com os parâmetros ajustados a partir dos dados de treino
    X_test = scaler.transform(X_test)

    return X_train, y_train, X_test, y_test


def diamond_preprocessing():
    df_diamonds = pd.read_csv("./experiments/diamonds.csv")
    df_diamonds = df_diamonds.drop(["Unnamed: 0"], axis=1)
    df_diamonds['cut'].replace({'Fair': 1, 'Good': 2, 'Very Good': 3, 'Premium': 4, 'Ideal': 5}, inplace=True)
    df_diamonds['color'].replace({'J': 1, 'I': 2, 'H': 3, 'G': 4, 'F': 5, 'E': 6, 'D': 7}, inplace=True)
    df_diamonds['clarity'].replace({'I1': 1, 'SI2': 2, 'SI1': 3, 'VS2': 4, 'VS1': 5, 'VVS2': 6, 'VVS1': 7, 'IF': 8},
                                   inplace=True)
    # Divisão entre treino e teste
    X = df_diamonds.drop(['price'], 1)
    y = df_diamonds['price']
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2)
    scaler = StandardScaler()
    X_train = scaler.fit_transform(X_train)
    X_test = scaler.transform(X_test)

    return X_train, y_train, X_test, y_test


def run(data, algorithm: BaseEstimator, hyperopt_method, is_hyperbrkga: bool, parameters: dict, filename: str,
        df: DataFrame):
    X_train, y_train, X_test, y_test = data

    # Hyperparameter Optimization
    if is_hyperbrkga:
        hyper_opt = hyperopt_method(algorithm(), parameters=parameters, cv=5, data=X_train, target=y_train)
    else:
        hyper_opt = hyperopt_method(algorithm(), parameters, cv=5)

    hyper_opt.fit(X_train, y_train)

    return get_metrics(filename, hyper_opt, is_hyperbrkga, df)


if __name__ == "__main__":
    classifier_algorithms = [
        LogisticRegression,
        DecisionTreeClassifier,
        RandomForestClassifier,
        KNeighborsClassifier,
        XGBClassifier,
        MLPClassifier
    ]
    regression_algorithms = [
        LinearRegression,
        Lasso,
        DecisionTreeRegressor,
        RandomForestRegressor,
        XGBRegressor,
        MLPRegressor,
    ]

    classifier_hyperparams = {
        "LogisticRegression": {
            'penalty': ('l1', 'l2', 'elasticnet', 'none'),
            'solver': ('newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga'),
            'C': [0.001, 0.01, 0.1, 1, 10, 100, 1000]
        },
        "DecisionTreeClassifier": {
            'criterion': ('gini', 'entropy', 'log_loss'),
            'splitter': ('best', 'random'),
            'max_depth': np.arange(0, 64),
            'min_samples_split': np.arange(0, 20),
            'min_samples_leaf': np.arange(0, 20)
        },
        "RandomForestClassifier": {
            'criterion': ('gini', 'entropy', 'log_loss'),
            'max_depth': np.arange(0, 64),
            'min_samples_split': np.arange(0, 20),
            'min_samples_leaf': np.arange(0, 20)
        },
        "KNeighborsClassifier": {
            'n_neighbors': np.arange(1, 30),
            'leaf_size': np.arange(1, 50),
            'algorithm': ('auto', 'ball_tree', 'kd_tree', 'brute'),
            'p': [1, 2]
        },
        "XGBClassifier": {
            'max_depth': np.arange(0, 64),
            'max_leaves': np.arange(0, 64),
            'learning_rate': [0.001, 0.01, 0.125, 0.25, 0.5, 1],
            'grow_policy': ('depthwise', 'lossguide')
        },
        "MLPClassifier": {
            'hidden_layer_sizes': [(10, 30, 10), (20,)],
            'activation': ['tanh', 'relu'],
            'solver': ['sgd', 'adam'],
            'alpha': [0.0001, 0.05],
            'learning_rate': ['constant', 'adaptive']
        },
    }

    regression_hyperparams = {
        "LinearRegression": {
            'fit_intercept': [True, False],
            'positive': [True, False]
        },
        "Lasso": {
            'alpha': np.linspace(0.0, 10.0),
            'fit_intercept': [True, False],
            'selection': ('cyclic', 'random')
        },
        "DecisionTreeRegressor": {
            'criterion': ('squared_error', 'friedman_mse', 'absolute_error', 'poisson'),
            'splitter': ('best', 'random'),
            'max_depth': np.arange(0, 20),
            'min_samples_split': np.arange(0, 20),
            'min_samples_leaf': np.arange(0, 20)
        },
        "RandomForestRegressor": {
            'criterion': ('squared_error', 'friedman_mse', 'absolute_error', 'poisson'),
            'max_depth': np.arange(0, 20),
            'min_samples_split': np.arange(0, 20),
            'min_samples_leaf': np.arange(0, 20)
        },
        "XGBRegressor": {
            'max_depth': np.arange(0, 30),
            'max_leaves': np.arange(0, 30),
            'learning_rate': [0.001, 0.01, 0.125, 0.25, 0.5, 1],
            'n_estimators': np.arange(1, 10000, 500)
        },
        "MLPRegressor": {
            'hidden_layer_sizes': [(10, 30, 10), (20,)],
            'activation': ['tanh', 'relu'],
            'solver': ['sgd', 'adam'],
            'alpha': np.linspace(0.0001, 0.05, 20),
            'learning_rate': ['constant', 'adaptive']
        },
    }

    hyperopt_methods = [HyperBRKGASearchCV, GridSearchCV, RandomizedSearchCV]

    data_cred = cred_preprocessing()
    data_diamond = diamond_preprocessing()

    i = 0

    while i < len(classifier_algorithms):
        algorithm = classifier_algorithms[i]
        params = classifier_hyperparams[algorithm.__name__]
        for k in range(len(hyperopt_methods)):
            hom = hyperopt_methods[k]
            dataframe = DataFrame()
            filename = f"experiments/results/{algorithm.__name__}_{hom.__name__}_cred.csv"
            file = open(filename, 'a')

            for x in range(10):
                dataframe, metrics = run(data_cred, algorithm, hom, k == 0, params, filename, dataframe)
                file.write(f'{x},{metrics["Score"]},{metrics["Time"]},{str(metrics["Param"])}\n')

            file.close()
            compute_mean_and_std(dataframe, filename)
        i += 1

    j = 0

    while j < len(regression_algorithms):
        algorithm = regression_algorithms[j]
        params = regression_hyperparams[algorithm.__name__]
        for k in range(len(hyperopt_methods)):
            hom = hyperopt_methods[k]
            dataframe = DataFrame()
            filename = f"experiments/results/{algorithm.__name__}_{hom.__name__}_diamond.csv"
            file = open(filename, 'a')

            for x in range(5):
                dataframe, metrics = run(data_diamond, algorithm, hom, k == 0, params, filename, dataframe)
                file.write(f'{x},{metrics["Score"]},{metrics["Time"]},{str(metrics["Param"])}\n')

            file.close()
            compute_mean_and_std(dataframe, filename)
        j += 1



import copy
import random
from datetime import datetime

import numpy as np
from sklearn import clone
from sklearn.base import is_classifier
from sklearn.metrics import check_scoring
from sklearn.metrics._scorer import _check_multimetric_scoring
from sklearn.model_selection import check_cv
from sklearn.model_selection._search import BaseSearchCV, ParameterGrid
from sklearn.model_selection._validation import cross_val_score
from sklearn.utils import indexable
from sklearn.utils.validation import _check_fit_params

from skga.hbrkga import BrkgaMpIpr, Sense, BaseChromosome, load_configuration_from_dict, BayesianOptimizerElites

class Decoder:

    def __init__(self, parameters, estimator, X, y, cv):
        self._parameters = parameters
        self._estimator = estimator
        self._X = X
        self._y = y
        self._limits = [self._parameters[l] for l in list(self._parameters.keys())]
        self._cv = cv

    def decode(self, chromosome: BaseChromosome, rewrite: bool) -> float:
        return self.score(self.encoder(chromosome))

    def encoder(self, chromosome: BaseChromosome) -> dict:
        chr_size = len(chromosome)
        hyperparameters = copy.deepcopy(self._parameters)

        for geneIdx in range(chr_size):
            gene = chromosome[geneIdx]
            key = list(self._parameters.keys())[geneIdx]
            limits = self._parameters[key]  # evita for's aninhados
            if type(limits) is np.ndarray:
                limits = limits.tolist()

            if type(limits[0]) is str:
                hyperparameters[key] = limits[round(gene * (len(limits) - 1))]
            elif type(limits[0]) is int and len(limits) > 2:
                hyperparameters[key] = int(limits[round(gene * (len(limits) - 1))])
            elif type(limits[0]) is bool:
                hyperparameters[key] = 1 if limits[0] else 0
            else:
                hyperparameters[key] = (gene * (limits[1] - limits[0])) + limits[0]

        return hyperparameters

    def score(self, hyperparameters: dict) -> float:
        estimator_clone = clone(self._estimator)
        estimator_clone.set_params(**hyperparameters)

        try:
            estimator_clone.fit(self._X, self._y)
        except ValueError:
            return 0.0

        # Adicionar o parâmetro scoring para que siga fielmente o SKLearn
        return cross_val_score(estimator_clone, self._X, self._y, cv=self._cv).mean()


class HyperBRKGASearchCV(BaseSearchCV):

    def __init__(
            self,
            estimator,
            brkga_params=None,
            exploit_method=None,
            *,
            scoring=None,
            n_jobs=None,
            refit=True,
            cv=None,
            verbose=0,
            pre_dispatch="2*n_jobs",
            error_score=np.nan,
            return_train_score=True,
            parameters,
            data,
            target
    ):
        super().__init__(
            estimator=estimator,
            scoring=scoring,
            n_jobs=n_jobs,
            refit=refit,
            cv=cv,
            verbose=verbose,
            pre_dispatch=pre_dispatch,
            error_score=error_score,
            return_train_score=return_train_score,
        )
        if brkga_params is not None:
            self.brkga_config, _ = brkga_params
        else:
            self.brkga_config , _ = load_configuration_from_dict({
                'population_size': 10,
                'elite_percentage': 0.3,
                'mutants_percentage': 0.3,
                'num_elite_parents': 1,
                'total_parents': 2,
                'bias_type': 'LOGINVERSE',
                'num_independent_populations': 1,
                'pr_number_pairs': 0,
                'pr_minimum_distance': 0.15,
                'pr_type': 'PERMUTATION',
                'pr_selection': 'BESTSOLUTION',
                'alpha_block_size': 1.0,
                'pr_percentage': 1.0,
                'exchange_interval': 0,
                'num_exchange_indivuduals': 0,
                'reset_interval': 0
            })
        self._parameters = parameters

        self.decoder = Decoder(self._parameters, estimator, data, target, cv)
        elite_number = int(self.brkga_config.elite_percentage * self.brkga_config.population_size)

        if exploit_method is not None:
            self.em = exploit_method(self.decoder, percentage=0.6, eliteNumber=elite_number)
        else:
            self.em = BayesianOptimizerElites(decoder=self.decoder, e=0.3, steps=3, percentage=0.6,
                                              eliteNumber=elite_number)

        chromosome_size = len(self._parameters)
        self.brkga = BrkgaMpIpr(
            decoder=self.decoder,
            sense=Sense.MAXIMIZE,
            seed=random.randint(-10000, 10000),
            chromosome_size=chromosome_size,
            params=self.brkga_config,
            diversity_control_on=True,
            n_close=3,
            exploitation_method=self.em
        )

        self.brkga.initialize()

    def fit(self, X, y=None, *, groups=None, **fit_params):
        estimator = self.estimator

        if callable(self.scoring):
            scorers = self.scoring
        elif self.scoring is None or isinstance(self.scoring, str):
            scorers = check_scoring(self.estimator, self.scoring)
        else:
            scorers = _check_multimetric_scoring(self.estimator, self.scoring)
            self._check_refit_for_multimetric(scorers)

        X, y, groups = indexable(X, y, groups)
        fit_params = _check_fit_params(X, fit_params)

        cv_orig = check_cv(self.cv, y, classifier=is_classifier(estimator))
        n_splits = cv_orig.get_n_splits(X, y, groups)

        def evaluate_candidates(candidate_params, cv=None, more_results=None):
            start = datetime.now()
            candidate_params = list(candidate_params)
            all_candidate_params = []

            for i in range(1, 11):
                print("\n###############################################")
                print(f"Generation {i}")
                print("")
                self.brkga.evolve()

                for pop_idx in range(len(self.brkga._current_populations)):
                    pop_diversity_score = self.brkga.calculate_population_diversity(pop_idx)
                    if self.verbose > 2:
                        print(f"Population {pop_idx}:")
                        print(f"Population diversity score = {pop_diversity_score}")
                        print("")
                        print("Chromosomes = ")
                        for chromo_idx in range(len(self.brkga._current_populations[pop_idx].chromosomes)):
                            print(f"{chromo_idx} -> {self.brkga._current_populations[pop_idx].chromosomes[chromo_idx]}")
                        print("")
                        print("Fitness = ")
                        for fitness in self.brkga._current_populations[pop_idx].fitness:
                            print(fitness)
                        print("------------------------------")

                best_cost = self.brkga.get_best_fitness()
                best_chr = self.brkga.get_best_chromosome()
                if self.verbose > 2:
                    print(f"{datetime.now()} - Best score so far: {best_cost}")
                    print(f"{datetime.now()} - Best chromosome so far: {best_chr}")
                    print(f"{datetime.now()} - Total time so far: {datetime.now() - start}", flush=True)

            best_cost = self.brkga.get_best_fitness()
            best_chr = self.brkga.get_best_chromosome()
            if self.verbose > 2:
                print("\n###############################################")
                print("Final results:")
                print(f"{datetime.now()} - Best score: {best_cost}")
                print(f"{datetime.now()} - Best chromosome: {best_chr}")
                print(f"Total time = {datetime.now() - start}")

            all_candidate_params.extend(candidate_params)
            self.results = {
                "best_chromosome": best_chr,
                "best_param_decoded": self.decoder.encoder(best_chr),
                "best_param_score": best_cost,
                "total_time": (datetime.now() - start).total_seconds(),
            }

        self._run_search(evaluate_candidates)

        # Store the only scorer not as a dict for single metric evaluation
        self.scorer_ = scorers

        self.cv_results_ = self.results
        self.n_splits_ = n_splits

        return self

    def _run_search(self, evaluate_candidates):
        evaluate_candidates(ParameterGrid(self._parameters))



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="skga",
    version="0.0.6",
    description="The python package implementing the HyperBRKGA algorithm optimizes hyperparameters of machine learning algorithms through a hybrid approach based on genetic algorithms.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MLRG-CEFET-RJ/skga",
    author="Leonardo Sauberman, João Pedro Nogueira, Eduardo Bezerra",
    author_email="leonardo.moraes@aluno.cefet-rj.br, joao.carneiro@aluno.cefet-rj.br, ebezerra@cefet-rj.br",
    license="BSD License",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    packages=["skga"],
    include_package_data=True,
    install_requires=[
        "bayesian-optimization==1.2.0",
        "bleach==6.0.0",
        "build==0.10.0",
        "certifi==2023.5.7",
        "charset-normalizer==3.1.0",
        "contourpy==1.0.7",
        "cycler==0.11.0",
        "Cython==0.29.30",
        "docutils==0.19",
        "fonttools==4.39.3",
        "idna==3.4",
        "importlib-metadata==6.6.0",
        "importlib-resources==5.12.0",
        "jaraco.classes==3.2.3",
        "joblib==1.2.0",
        "keyring==23.13.1",
        "kiwisolver==1.4.4",
        "markdown-it-py==2.2.0",
        "matplotlib==3.7.1",
        "mdurl==0.1.2",
        "more-itertools==9.1.0",
        "numpy==1.23.1",
        "packaging==23.1",
        "pandas==1.4.3",
        "Pillow==9.5.0",
        "pkginfo==1.9.6",
        "Pygments==2.15.1",
        "pyparsing==3.0.9",
        "python-dateutil==2.8.2",
        "pyproject_hooks==1.0.0",
        "pytz==2023.3",
        "readme-renderer==37.3",
        "requests==2.30.0",
        "requests-toolbelt==1.0.0",
        "rfc3986==2.0.0",
        "rich==13.3.5",
        "scikit-learn==1.1.2",
        "scipy==1.8.1",
        "seaborn==0.12.1",
        "six==1.16.0",
        "threadpoolctl==3.1.0",
        "tomli==2.0.1",
        "torch==1.9.0",
        "twine==4.0.2",
        "urllib3==1.26.15",
        "webencodings==0.5.1",
        "wheel==0.40.0",
        "xgboost==1.7.1",
        "zipp==3.15.0",
    ],
)